"use strict";
/// <reference types="codemirror/addon/hint/show-hint" />
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.hideCompletions = exports.showTriggeredCompletions = exports.showInvokedCompletions = void 0;
const vscode_languageserver_protocol_1 = require("vscode-languageserver-protocol");
const codemirror_1 = __importDefault(require("codemirror"));
const conversions_1 = require("../utils/conversions");
const editor_1 = require("../utils/editor");
const snippet_1 = require("./snippet");
/**
 * Show completion triggered by typing an identifier or manual invocation.
 * @param editor
 * @param items - Completion items from the server.
 * @param tokenRange - Range describing the span of the typed identifier.
 */
exports.showInvokedCompletions = (editor, items, [tokenFrom, tokenTo], renderMarkdown = (x) => x) => {
    if (items.length === 0)
        return;
    const typed = editor.getRange(tokenFrom, tokenTo);
    const filtered = filteredItems(items, typed, 30);
    showCompletionItems(editor, filtered, tokenFrom, tokenTo, renderMarkdown);
};
/**
 * Show completions triggered by a trigger character specified by the `triggerCharacters`.
 * Completion items are inserted differently from when it's invoked.
 * @param editor - The editor.
 * @param items - Completion items from the server.
 * @param pos - Current cursor position where the item will be inserted after.
 */
exports.showTriggeredCompletions = (editor, items, pos, renderMarkdown = (x) => x) => {
    if (items.length === 0)
        return;
    showCompletionItems(editor, items, pos, pos, renderMarkdown);
};
/**
 * Hide completion popup.
 * @param editor
 */
exports.hideCompletions = (editor) => {
    if (editor.state.completionActive)
        editor.state.completionActive.close();
};
const showCompletionItems = (editor, items, posFrom, posTo, renderMarkdown) => {
    editor.showHint({
        completeSingle: false,
        hint: () => withItemTooltip({
            from: posFrom,
            to: posTo,
            list: items.map((item) => {
                const displayText = item.label;
                let text = item.insertText || item.label;
                let from = posFrom;
                let to = posTo;
                if (item.textEdit) {
                    const edit = item.textEdit;
                    [from, to] = conversions_1.cmRange(vscode_languageserver_protocol_1.TextEdit.is(edit) ? edit.range : edit.replace);
                    text = edit.newText;
                }
                const isSnippet = item.insertTextFormat === vscode_languageserver_protocol_1.InsertTextFormat.Snippet;
                return {
                    // Note that `text` field is only set because @types/codemirror wrongly requires it.
                    text: "",
                    displayText,
                    hint: (cm) => {
                        if (isSnippet) {
                            // Insert snippet and start snippet mode.
                            return snippet_1.insertSnippet(cm, text, from, to, item.additionalTextEdits);
                        }
                        cm.replaceRange(text, from, to, "complete");
                        if (item.additionalTextEdits) {
                            editor_1.applyEdits(cm, item.additionalTextEdits, "+complete");
                        }
                    },
                    render: itemRenderer(item),
                    data: getItemData(item, renderMarkdown),
                };
            }),
        }),
    });
};
// Returns `hints` with an ability to show details of each completion item on `select`.
// See https://github.com/codemirror/CodeMirror/blob/6fcc49d5321261b525e50f111f3b28a602f01f71/addon/tern/tern.js#L229
const withItemTooltip = (hints) => {
    let tooltip = null;
    codemirror_1.default.on(hints, "close", () => {
        if (tooltip) {
            tooltip.remove();
            tooltip = null;
        }
    });
    codemirror_1.default.on(hints, "update", () => {
        if (tooltip) {
            tooltip.remove();
            tooltip = null;
        }
    });
    // @ts-ignore: @types/codemirror has (cur: Hint, node: Element)
    codemirror_1.default.on(hints, "select", (cur, node) => {
        tooltip === null || tooltip === void 0 ? void 0 : tooltip.remove();
        const data = cur.data;
        if (!data)
            return;
        let content;
        if (data.detail) {
            content = data.detail;
            if (data.documentation)
                content += "<br>" + data.documentation;
        }
        else {
            content = data.documentation;
        }
        if (!content)
            return;
        const x = node.parentNode.getBoundingClientRect().right +
            window.pageXOffset;
        const y = node.getBoundingClientRect().top + window.pageYOffset;
        const el = document.createElement("div");
        el.innerHTML = content;
        tooltip = document.createElement("div");
        tooltip.classList.add("cmw-completion-item-doc");
        tooltip.style.position = "absolute";
        tooltip.style.zIndex = "10";
        tooltip.style.left = `${x}px`;
        tooltip.style.top = `${y}px`;
        tooltip.style.fontSize = "12px";
        tooltip.style.padding = "2px";
        tooltip.style.maxHeight = "400px";
        tooltip.style.overflowY = "auto";
        tooltip.appendChild(el);
        document.body.appendChild(tooltip);
    });
    return hints;
};
const getItemData = (item, renderMarkdown) => ({
    kind: conversions_1.completionItemKindToString(item.kind),
    detail: item.detail || "",
    documentation: conversions_1.documentationToString(item.documentation, renderMarkdown),
});
// Returns a function to render `item` in completion popup.
const itemRenderer = (item) => (el) => {
    // TODO Use meaningful icons for each kind
    // TODO Show matching characters in different color?
    const hue = Math.floor(((item.kind || 0) / 30) * 180 + 180);
    let color = `hsl(${hue}, 75%, 50%)`;
    // If completing a color, show the color.
    // TODO This is just an example and only handles 6 digit hex.
    if (item.kind === vscode_languageserver_protocol_1.CompletionItemKind.Color) {
        const doc = conversions_1.documentationToString(item.documentation);
        if (doc) {
            const hex = doc.match(/#[\da-f]{6}/i);
            if (hex)
                color = hex[0];
        }
    }
    const icon = el.appendChild(document.createElement("div"));
    icon.style.color = color;
    icon.className = [
        `cmw-icon`,
        `cmw-icon--${conversions_1.completionItemKindToString(item.kind)}`,
    ].join(" ");
    el.appendChild(document.createTextNode(item.label));
};
const filteredItems = (items, typed, maxItems = 30) => {
    const scores = items.reduce((o, item) => {
        o[item.label] = lcsScore(item.filterText || item.label, typed);
        // Boost the score if preselect
        // if (item.preselect) o[item.label] *= 100;
        return o;
    }, {});
    return items
        .slice()
        .sort((a, b) => scores[b.label] - scores[a.label])
        .slice(0, maxItems);
};
// Similarity score between strings based on the length of the longest common subsequence.
const lcsScore = (item, typed) => {
    if (item.length === 0 || typed.length === 0)
        return 0;
    // Prefer shorter item if it includes typed word. Boost if it starts with it.
    if (item.includes(typed)) {
        return typed.length + 1 / item.length + (item.startsWith(typed) ? 1 : 0);
    }
    const lcs = item.length < typed.length
        ? lcsLength(typed, item)
        : lcsLength(item, typed);
    // Boost if the first character matches.
    if (item[0].toLowerCase() === typed[0].toLowerCase()) {
        // Another boost if matching case.
        return lcs + 1 + (item[0] === typed[0] ? 1 : 0);
    }
    return lcs;
};
// Returns LCS length.
// Uses single array of size min(x.length, y.length) + 1 for look up.
// Assumes x.length >= y.length > 0.
//
// For X = AGCAT, Y = GAC, L changes like the following:
//
//      Ø  G  A  C
//   Ø  0  0  0  0
//   A  0  0 *1  1
//   G  0 *1  1  1
//   C  0  1  1 *2
//   A  0  1 *2  2
//   T  0  1  2  2
//
// where * denotes a match that incremented the value from previous column/row.
const lcsLength = (x, y) => {
    const m = x.length;
    const n = y.length;
    const L = [0];
    for (let i = 1; i <= n; ++i)
        L[i] = 0;
    for (let i = 1; i <= m; ++i) {
        // prev is the value from previous column/row
        for (let j = 1, prev = L[0]; j <= n; ++j) {
            const tmp = L[j];
            L[j] = x[i - 1] === y[j - 1] ? prev + 1 : Math.max(L[j - 1], L[j]);
            prev = tmp;
        }
    }
    return L[n];
};
//# sourceMappingURL=completion.js.map