/// <reference types="codemirror/addon/hint/show-hint" />
import type { Editor } from "codemirror";
/**
 * Describes text document's language association.
 */
export interface LanguageAssociation {
    /**
     * Language ID associated with the file.
     */
    languageId: string;
    /**
     * IDs of language servers to connect to.
     * Accepts multiple IDs for future extension, but currently only the first one is used.
     */
    languageServerIds: string[];
}
/**
 * Options for Workspace.
 */
export interface WorkspaceOptions {
    /**
     * The URI of the project root.
     */
    rootUri: string;
    /**
     * Provide Language Server connection string.
     *
     * The returned string can be either:
     *
     * - URI of a WebSocket proxy of a Language Server (`wss?://`)
     * - Path to a script to start Language Server in Web Worker
     *
     * If the returned string does not start with `wss?://`, it's assumed to be a Worker script.
     */
    getConnectionString: (this: void, langserverId: string) => Promise<string>;
    /**
     * Provide language association (language id, language server id) for a file with the uri.
     */
    getLanguageAssociation: (this: void, uri: string) => LanguageAssociation | null;
    /**
     * Function to render Markdown to HTML string.
     * If not provided and the server's response contains Markdown, it'll be displayed as is.
     */
    renderMarkdown?: (this: void, markdown: string) => string;
}
/**
 * Workspace provides code intelligence to CodeMirror editors by managing
 * communications with Language Servers and adding event handlers.
 */
export declare class Workspace {
    private editors;
    private connectPromises;
    private connections;
    private documentVersions;
    private subscriptionDisposers;
    private getConnectionString;
    private getLanguageAssociation;
    private renderMarkdown;
    private canHandleMarkdown;
    private rootUri;
    /**
     * Create new workspace.
     * @param options
     */
    constructor(options: WorkspaceOptions);
    /**
     * Dispose the workspace.
     * Close connections and remove references to editors.
     */
    dispose(): void;
    /**
     * Open text document in the workspace to notify the Language Server and
     * enable code intelligence.
     * @param path - The file path relative to the project root.
     * @param editor - CodeMirror Editor instance.
     */
    openTextDocument(path: string, editor: Editor): Promise<void>;
    private addEventHandlers;
    /**
     * Close text document in the workspace to notify the Language Server and
     * remove everything added by the workspace.
     * @param path - The file path relative to the project root.
     */
    closeTextDocument(path: string): Promise<void>;
    /**
     * Notify the Language Server that the text document was saved.
     *
     * If connected to a WebSocket proxy with synchronization enabled,
     * the contents of the file will be written to disk.
     * @param path - The file path relative to the project root.
     */
    saveTextDocument(path: string): Promise<void>;
    private removeEventHandlers;
    /**
     * Private method to connect to the language server if possible.
     * If existing connection exists, it'll be shared.
     *
     * @param serverId - ID of the language server.
     */
    private connect;
    private getDocumentUri;
}
